import functools
from typing import Any
from typing import Callable
from typing import Dict
from typing import List
from typing import Tuple


class InstanceCreationError(Exception):
    pass


class Constant(type):
    __dunder_exclude_list = [
        "__dict__",
        "__weakref__",
        "__doc__",
        "__metaclass__",
        "__wrapped__",
        "__qualname__",
        "__module__",
    ]

    def __new__(cls, name: str, bases: Tuple, arg: Dict):
        def _raise_error(x):
            raise InstanceCreationError(
                f"Cannot create an instance of {x.__class__.__name__}"
            )

        def __select_fields(mapping: Dict, exclude_list: List = []):
            return {key: mapping[key] for key in mapping if key not in exclude_list}

        inst = super().__new__(
            cls,
            name,
            bases,
            {
                "_Constant__frozen": False,
                "_members": __select_fields(
                    arg, exclude_list=Constant.__dunder_exclude_list
                ),
                **arg,
            },
        )
        inst.__init__ = _raise_error
        inst.__frozen = True
        return inst

    def __setattr__(self, key: str, value: Any):
        if self.__frozen:
            raise AttributeError("Cannot set or change the class attributes")
        super().__setattr__(key, value)

    def __str__(cls):
        return cls.__name__

    def __repr__(cls):
        return cls.__name__


def constant(cls) -> Constant:
    __name = str(cls.__name__)
    __bases = tuple(cls.__bases__)
    __dict = dict(cls.__dict__)

    for each_slot in __dict.get("__slots__", tuple()):
        __dict.pop(each_slot, None)

    __dict["__metaclass__"] = Constant
    __dict["__wrapped__"] = cls
    return Constant(__name, __bases, __dict)


def dispatch(func: Callable) -> Callable:
    dispatch_map = {}

    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        key = args[0]
        if key in dispatch_map:
            return dispatch_map[key](*args[1:], **kwargs)
        return func(*args, **kwargs)

    def register(key):
        def decorator(func_):
            dispatch_map[key] = func_
            return func_

        return decorator

    wrapper.register = register
    return wrapper



# -*- coding: utf-8 -*-
from setuptools import setup

modules = \
['sthira']
setup_kwargs = {
    'name': 'sthira',
    'version': '0.2.0',
    'description': 'The `Constant` class is a metaclass for creating classes with constant attributes',
    'long_description': '# Sthira\n\n> The word for "constant" in Sanskrit can be translated as "स्थिर" (sthira)\n\n\nThe `Constant` class is a metaclass for creating classes with constant attributes.\nOnce set, the attributes of a `Constant` class cannot be changed, and new attributes cannot be added.\nThis allows for creating classes that represent unchangeable values, such as constants, enums, and similar constructs.\nThe Constant class also provides a `__str__`and `__repr__` implementation for convenient representation of the class.\n\n# Installtion\n\n`pip install sthira`\n\n## Usage\n\n```python\nfrom sthira import constant\n\n\n@constant\nclass Mammal:\n    HUMAN = "human"\n    TIGER = "tiger"\n    LION = "lion"\n\n@constant\nclass Bird:\n    CROW = "crow"\n    HAWK = "hawk"\n\n@constant\nclass Fish:\n    TUNA = "tuna"\n\n@constant\nclass Animal:\n    MAMMAL = Mammal\n    BIRD = Bird\n    FISH = Fish\n\nprint(f"{Animal.MAMMAL}")\nprint(f"{Animal.MAMMAL.HUMAN}")\nprint(f"{Animal.BIRD.CROW}")\n\n```\n\n> Output\n\n```\nMammal\nhuman\ncrow\n```\n\n## Cannot modify attributes\n\n```python\nAnimal.MAMMAL.HUMAN = "HomoSapiens"\n\n#     raise AttributeError("Cannot set or change the class attributes")\n# AttributeError: Cannot set or change the class attributes\n```\n\n## Dispatch\n\n> check `test_constant.py`\n\n```python\nfrom sthira import dispatch\n\n@constant\nclass Red:\n    BRICK = "#AA4A44"\n    CADMIUM = "##D22B2B"\n\n\n@constant\nclass Green:\n    LIME = "#32CD32"\n    LIGHT = "#90EE90"\n\n\n@constant\nclass Color:\n    RED = Red\n    GREEN = Green\n    YELLOW = "not_there_yet"\n\n@dispatch\ndef get_color(color, input_):\n    # Default implementation\n    raise NotImplementedError("Unsupported color!")\n\n@get_color.register(Color.RED)\ndef _(input_):\n    return "I\'m red"\n\n@get_color.register(Color.GREEN)\ndef _(input_):\n    return "hulk out!"\n\nprint(get_color(Color.GREEN, "input"))\n# hulk out!\n```\n\n\n## Unit tests\n\n```python -m unittest test_constant.py```\n\n',
    'author': 'neelabalan',
    'author_email': 'neelabalan.n@outlook.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'None',
    'py_modules': modules,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)
